/*****************************************************************************
 *
 * utouch-grail - Multitouch Gesture Recognition Library
 *
 * Copyright (C) 2011 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#ifndef UTOUCH_GRAIL_RECOGNIZER_H_
#define UTOUCH_GRAIL_RECOGNIZER_H_

#include <list>
#include <map>
#include <memory>
#include <set>

#include <utouch/frame.h>

#include "utouch/grail.h"
#include "v3/forward.h"
#include "v3/subscription.h"

#define INSERT_TOUCH(element, set) \
  { \
    (set).insert(element); \
    LOG(Dbg) << "touch " << element << " has been added to " #set "\n"; \
  }

#define ERASE_TOUCH(element, set) \
  { \
    (set).erase(element); \
    LOG(Dbg) << "touch " << element << " has been erased from " #set "\n"; \
  }

/* OBS: it avoids the "expensive" ToString() call when debug output is not
   wanted. */
#define CLEAR_TOUCHES(set) \
  { \
    if (utouch::grail::Logger::instance().level() <= utouch::grail::Logger::Dbg) \
      if ((set).size() > 0) \
        LOG(Dbg) << "touch(es) " << (set).ToString() \
                 << " have been erased from " #set "\n"; \
    (set).clear(); \
  }

namespace utouch {
namespace grail {

class Recognizer {
 public:
  Recognizer(UGHandle* handle, const UFDevice device, UFWindowId window);
  virtual ~Recognizer();

  virtual bool atomic() const = 0;
  virtual void ProcessFrameEvent(const UFEvent event) = 0;

  unsigned int num_subscriptions() const {return num_subscriptions_;}

  UGStatus ActivateSubscription(UGSubscription* subscription);
  void DeactivateSubscription(UGSubscription* subscription);
  virtual void UpdateTime(uint64_t time);
  virtual uint64_t NextTimeout();
  UGStatus AcceptGesture(unsigned int id);
  UGStatus RejectGesture(unsigned int id);

  UGHandle* handle() const { return handle_; }
  float device_x_res() const { return device_x_res_; }
  float device_y_res() const { return device_y_res_; }
  bool device_direct() const { return device_direct_; }

  Recognizer(const Recognizer&) = delete;
  Recognizer& operator=(const Recognizer&) = delete;

 protected:
  static uint64_t kCompositionTime; /* in milliseconds */

  UGHandle* const handle_;
  const UFDevice device_;
  const UFWindowId window_id_;
  float device_x_res_; /* Units of pixel/mm */
  float device_y_res_;
  bool device_direct_;
  bool atomic_;
  std::set<UGSubscription*> subscriptions_[5];
  std::set<SharedGesture> unaccepted_gestures_;
  std::set<SharedGesture> accepted_gestures_;
  std::map<UFTouchId, uint64_t> start_times_;
  TouchSet unaccepted_touches_;
  TouchSet accepted_touches_;
  TouchSet free_touches_;

  unsigned int num_subscriptions_;

  void CheckConstructionFinished(uint64_t time);
  void RejectGesture(SharedGesture gesture);
  void ProcessEvent(const UFEvent& event);
};

} // namespace grail
} // namespace utouch
#endif // UTOUCH_GRAIL_RECOGNIZER_H_
