/*****************************************************************************
 *
 * grail - Gesture Recognition And Instantiation Library
 *
 * Copyright (C) 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "v3/regular-recognizer.h"

#include <algorithm>
#include <cassert>
#include <cstdint>
#include <cstdio>
#include <limits>

#include <utouch/frame.h>
#include <utouch/frame_x11.h>

#include "v3/handle.h"
#include "v3/gesture.h"
#include "v3/log.h"

namespace utouch {
namespace grail {

/**
 * @internal
 * Create a new regular recognizer for a given device and window
 */
RegularRecognizer::RegularRecognizer(UGHandle* handle, const UFDevice device,
                                     UFWindowId window_id)
    : Recognizer(handle, device, window_id) {
}

/**
 * @internal
 * Process a uTouch-Frame event
 */
void RegularRecognizer::ProcessFrameEvent(const UFEvent event) {
  LOG(Dbg) << "new event " << event << " with time "
           << frame_event_get_time(event) << "\n";
  UpdateTime(frame_event_get_time(event));
  MatchSubscriptionsForEvent(event);
  ProcessEvent(event);
}

/**
 * @internal
 * Analyze a uTouch-Frame event to see if any new gestures should begin
 */
void RegularRecognizer::MatchSubscriptionsForEvent(const UFEvent event) {
  /* Check if any subscriptions are active before doing any processing */
  if (num_subscriptions_ == 0)
    return;

  UFFrame frame;
  UFStatus status = frame_event_get_property(event, UFEventPropertyFrame,
                                             &frame);
  if (status != UFStatusSuccess) {
    LOG(Warn) << "failed to get frame from event\n";
    return;
  }

  /* Process all the touches that began in this frame */
  unsigned int num_touches = frame_frame_get_num_touches(frame);
  for (unsigned int i = 0; i < num_touches; ++i) {
    UFTouch touch;
    status = frame_frame_get_touch_by_index(frame, i, &touch);
    if (status != UFStatusSuccess) {
      LOG(Warn) << "failed to get touch from frame\n";
      continue;
    }

    if (frame_touch_get_state(touch) != UFTouchStateBegin)
      continue;

    UFTouchId touch_id = frame_touch_get_id(touch);

    /* Note touch start time and add to initial touch lists */
    start_times_[touch_id] = frame_touch_get_start_time(touch);
    LOG(Dbg) << "touch " << touch_id << " began with start time "
             << start_times_[touch_id] << "\n";

    INSERT_TOUCH(touch_id, unaccepted_touches_);
    INSERT_TOUCH(touch_id, free_touches_);

    /* HandleNewTouchForAcceptedGesture may erase the gesture from
     * accepted_gestures_, so we can't use range-based for loops */
    for (auto it = accepted_gestures_.begin();
         it != accepted_gestures_.end();
         ) {
      const SharedGesture& gesture = *it++;
      HandleNewTouchForAcceptedGesture(touch_id, gesture);
    }

    /* HandleNewTouchForAcceptedGesture may erase the gesture from
     * accepted_gestures_, so we can't use range-based for loops */
    for (auto it = unaccepted_gestures_.begin();
         it != unaccepted_gestures_.end();
         ) {
      const SharedGesture& gesture = *it++;
      HandleNewTouchForUnacceptedGesture(touch_id, gesture);
    }

    /* Attempt to match new gestures for active subscriptions */
    MatchOneTouchGestures(touch_id);
    MatchTwoTouchGestures(touch_id);
    MatchThreeTouchGestures(touch_id);
    MatchFourTouchGestures(touch_id);
    MatchFiveTouchGestures(touch_id);
  }
}

/**
 * @internal
 * Perform tasks necessary for when a new touch occurs and there is an existing
 * accepted gesture
 *
 * If a gesture may add a touch without crossing the maximum for the
 * subscription, create a new unaccepted gesture with the new touch.
 * Otherwise, do nothing. New gestures may still begin elsewhere.
 */
void RegularRecognizer::HandleNewTouchForAcceptedGesture(
    UFTouchId touch_id,
    const SharedGesture& gesture) {
  UGSubscription* subscription = gesture->subscription();
  if (gesture->current_touches().size() < subscription->touches_max()
      && !gesture->ContainsTouch(touch_id)) {
    TouchSet set(gesture->current_touches());
    set.insert(touch_id);
    Gesture* new_gesture = new Gesture(gesture.get(), set);
    LOG(Dbg) << "touch " << touch_id << " has been added to accepted gesture "
      << gesture->id() << " to create new gesture "
      << new_gesture->id() << "\n";
    unaccepted_gestures_.insert(SharedGesture(new_gesture));
    LOG(Dbg) << "gesture " << new_gesture << " has been added to unaccepted "
      "gestures\n";
  }
}

/**
 * @internal
 * Perform tasks necessary for when a new touch occurs and there is an existing
 * unaccepted gesture
 *
 * If a gesture may add a touch without crossing the maximum for the
 * subscription, create a new unaccepted gesture with the new touch.
 * Otherwise, do nothing. New gestures may still begin elsewhere.
 */
void RegularRecognizer::HandleNewTouchForUnacceptedGesture(
    UFTouchId touch_id,
    const SharedGesture& gesture) {
  UGSubscription* subscription = gesture->subscription();
  if (gesture->current_touches().size() < subscription->touches_max()
        && !gesture->ContainsTouch(touch_id)) {
    TouchSet set(gesture->current_touches());
    set.insert(touch_id);
    Gesture* new_gesture = new Gesture(gesture.get(), set);
    LOG(Dbg) << "touch " << touch_id
      << " has been added to unaccepted gesture " << gesture->id()
      << " to create new gesture " << new_gesture->id() << "\n";
    unaccepted_gestures_.insert(SharedGesture(new_gesture));
    LOG(Dbg) << "gesture " << new_gesture << " has been added to unaccepted "
      "gestures\n";
  }
}

/**
 * @internal
 * Attempt to match the given touch against one touch subscriptions
 */
void RegularRecognizer::MatchOneTouchGestures(UFTouchId touch_id) {
  for (UGSubscription* subscription : subscriptions_[0]) {
    TouchSet set;
    set.insert(touch_id);
    Gesture* gesture = new Gesture(this, subscription, set,
                                   start_times_[touch_id]);
    unaccepted_gestures_.insert(SharedGesture(gesture));

    LOG(Dbg) << "New tentative gesture " << gesture->id()
             << " matched subscription " << subscription << " with mask "
             << subscription->mask() << " for touch " << touch_id << "\n";
  }
}

/**
 * @internal
 * Attempt to match the given touch against two touch subscriptions
 */
void RegularRecognizer::MatchTwoTouchGestures(UFTouchId touch_id) {
  for (UGSubscription* subscription : subscriptions_[1]) {
    for (UFTouchId other_id : free_touches_) {
      if (other_id == touch_id)
        continue;

      /* All touches in a gesture must begin within a composition timeframe */
      uint64_t min_start_time = start_times_[touch_id];
      if (start_times_[other_id] < min_start_time &&
          unaccepted_touches_.find(other_id) != unaccepted_touches_.end())
        min_start_time = start_times_[other_id];

      if (start_times_[touch_id] - min_start_time < kCompositionTime) {
        TouchSet set;
        set.insert(touch_id);
        set.insert(other_id);
        Gesture* gesture = new Gesture(this, subscription, set,
                                       start_times_[touch_id]);
        unaccepted_gestures_.insert(SharedGesture(gesture));

        LOG(Dbg) << "New tentative gesture " << gesture->id()
                 << " matched subscription " << subscription << " with mask "
                 << subscription->mask() << " for touches " << touch_id << ", "
                 << other_id << "\n";
      }
    }
  }
}

/**
 * @internal
 * Attempt to match the given touch against three touch subscriptions
 */
void RegularRecognizer::MatchThreeTouchGestures(UFTouchId touch_id) {
  for (UGSubscription* subscription : subscriptions_[2]) {
    for (UFTouchId other_id_1 : free_touches_) {
      if (other_id_1 == touch_id)
        continue;

      for (UFTouchId other_id_2 : free_touches_) {
        if (other_id_2 <= other_id_1 || other_id_2 == touch_id)
          continue;

        /* All touches in a gesture must begin within a composition timeframe */
        uint64_t min_start_time = start_times_[touch_id];
        if (start_times_[other_id_1] < min_start_time &&
            unaccepted_touches_.find(other_id_1) != unaccepted_touches_.end())
          min_start_time = start_times_[other_id_1];
        if (start_times_[other_id_2] < min_start_time &&
            unaccepted_touches_.find(other_id_2) != unaccepted_touches_.end())
          min_start_time = start_times_[other_id_2];

        if (start_times_[touch_id] - min_start_time < kCompositionTime) {
          TouchSet set;
          set.insert(touch_id);
          set.insert(other_id_1);
          set.insert(other_id_2);
          Gesture* gesture = new Gesture(this, subscription, set,
                                         start_times_[touch_id]);
          unaccepted_gestures_.insert(SharedGesture(gesture));

          LOG(Dbg) << "New tentative gesture " << gesture->id()
                   << " matched subscription " << subscription << " with mask "
                   << subscription->mask() << " for touches " << touch_id
                   << ", " << other_id_1 << ", " << other_id_2 << "\n";
        }
      }
    }
  }
}

/**
 * @internal
 * Attempt to match the given touch against four touch subscriptions
 */
void RegularRecognizer::MatchFourTouchGestures(UFTouchId touch_id) {
  for (UGSubscription* subscription : subscriptions_[3]) {
    for (UFTouchId other_id_1 : free_touches_) {
      if (other_id_1 == touch_id)
        continue;

      for (UFTouchId other_id_2 : free_touches_) {
        if (other_id_2 <= other_id_1 || other_id_2 == touch_id)
          continue;

        for (UFTouchId other_id_3 : free_touches_) {
          if (other_id_3 <= other_id_2 || other_id_3 == touch_id)
            continue;

          /* All touches in a gesture must begin within a composition
           * timeframe */
          uint64_t min_start_time = start_times_[touch_id];
          if (start_times_[other_id_1] < min_start_time &&
              unaccepted_touches_.find(other_id_1) != unaccepted_touches_.end())
            min_start_time = start_times_[other_id_1];
          if (start_times_[other_id_2] < min_start_time &&
              unaccepted_touches_.find(other_id_2) != unaccepted_touches_.end())
            min_start_time = start_times_[other_id_2];
          if (start_times_[other_id_3] < min_start_time &&
              unaccepted_touches_.find(other_id_3) != unaccepted_touches_.end())
            min_start_time = start_times_[other_id_3];

          if (start_times_[touch_id] - min_start_time < kCompositionTime) {
            TouchSet set;
            set.insert(touch_id);
            set.insert(other_id_1);
            set.insert(other_id_2);
            set.insert(other_id_3);
            Gesture* gesture = new Gesture(this, subscription, set,
                                           start_times_[touch_id]);
            unaccepted_gestures_.insert(SharedGesture(gesture));

            LOG(Dbg) << "New tentative gesture " << gesture->id()
                     << " matched subscription " << subscription
                     << " with mask " << subscription->mask() << " for touches "
                     << touch_id << ", " << other_id_1 << ", " << other_id_2
                     << ", " << other_id_3 << "\n";
          }
        }
      }
    }
  }
}

/**
 * @internal
 * Attempt to match the given touch against five touch subscriptions
 */
void RegularRecognizer::MatchFiveTouchGestures(UFTouchId touch_id) {
  for (UGSubscription* subscription : subscriptions_[4]) {
    for (UFTouchId other_id_1 : free_touches_) {
      if (other_id_1 == touch_id)
        continue;

      for (UFTouchId other_id_2 : free_touches_) {
        if (other_id_2 <= other_id_1 || other_id_2 == touch_id)
          continue;

        for (UFTouchId other_id_3 : free_touches_) {
          if (other_id_3 <= other_id_2 || other_id_3 == touch_id)
            continue;

          for (UFTouchId other_id_4 : free_touches_) {
            if (other_id_4 <= other_id_3 || other_id_4 == touch_id)
              continue;

            /* All touches in a gesture must begin within a composition
             * timeframe */
            uint64_t min_start_time = start_times_[touch_id];
            if (start_times_[other_id_1] < min_start_time &&
                unaccepted_touches_.find(other_id_1) !=
                    unaccepted_touches_.end())
              min_start_time = start_times_[other_id_1];
            if (start_times_[other_id_2] < min_start_time &&
                unaccepted_touches_.find(other_id_2) !=
                    unaccepted_touches_.end())
              min_start_time = start_times_[other_id_2];
            if (start_times_[other_id_3] < min_start_time &&
                unaccepted_touches_.find(other_id_3) !=
                    unaccepted_touches_.end())
              min_start_time = start_times_[other_id_3];
            if (start_times_[other_id_4] < min_start_time &&
                unaccepted_touches_.find(other_id_4) !=
                    unaccepted_touches_.end())
              min_start_time = start_times_[other_id_4];

            if (start_times_[touch_id] - min_start_time < kCompositionTime) {
              TouchSet set;
              set.insert(touch_id);
              set.insert(other_id_1);
              set.insert(other_id_2);
              set.insert(other_id_3);
              set.insert(other_id_4);
              Gesture* gesture = new Gesture(this, subscription, set,
                                             start_times_[touch_id]);
              unaccepted_gestures_.insert(SharedGesture(gesture));

              LOG(Dbg) << "New tentative gesture " << gesture->id()
                       << " matched subscription " << subscription
                       << " with mask " << subscription->mask()
                       << " for touches " << touch_id << ", " << other_id_1
                       << ", " << other_id_2 << ", " << other_id_3 << ", "
                       << other_id_4 << "\n";
            }
          }
        }
      }
    }
  }
}

} // namespace grail
} // namespace utouch
